import flask

from ...usage import usage
from .main import MainController


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_forms", __name__)

    @bp.get("/")
    @usage
    def _get_forms():
        return [f.editor_dto for f in controller.get_forms()]

    @bp.get("/<path:path>")
    @usage
    def _get_form(path: str):
        form = controller.get_form(path)
        if not form:
            flask.abort(404)
        return form.editor_dto

    @bp.delete("/<path:path>")
    @usage
    def _delete_form(path: str):
        controller.delete_form(path)
        return {"success": True}

    @bp.post("/")
    @usage
    def _create_form():
        form = controller.create_form()
        return form.editor_dto

    @bp.put("/<path:path>")
    @usage
    def _update_form(path: str):
        data = flask.request.json
        if not data:
            flask.abort(400)

        form = controller.update_stage(path, data)
        return form.editor_dto if form else None

    return bp



from dataclasses import dataclass
import typing


class Message:
    type: str
    data: dict

    def __init__(self, data: dict):
        self.data = data

    def to_json(self, is_preview: bool):
        return {"type": self.type, **self.data}


class StdioMessage(Message):
    type = "stdio"

    def __init__(self, type: str, log: str):
        super().__init__({"log": log})
        self.type = type


class AuthRequireInfoMessage(Message):
    type = "auth:require-info"

    def __init__(self, refresh: bool):
        super().__init__({"refresh": refresh})


class AuthInvalidJWTMessage(Message):
    type = "auth:invalid-jwt"

    def __init__(self):
        super().__init__({})


class AuthValidJWTMessage(Message):
    type = "auth:valid-jwt"

    def __init__(self):
        super().__init__({})


class RedirectMessage(Message):
    type = "redirect"

    def __init__(self, url: str, query_params: dict):
        super().__init__({"url": url, "queryParams": query_params})


class AlertMessage(Message):
    type = "alert"

    def __init__(self, message: str, severity: str):
        super().__init__({"message": message, "severity": severity})


class ExecuteJSRequestMessage(Message):
    type = "execute-js:request"

    def __init__(self, code: str, context: dict):
        super().__init__({"code": code, "context": context})


class FilesChangedMessage(Message):
    type = "files:changed"

    def __init__(self):
        super().__init__({})


class LockFailedMessage(Message):
    type = "lock-failed"

    def __init__(self, status: typing.Optional[str]):
        super().__init__({"status": status})


@dataclass
class CloseDTO:
    exit_status: str
    exception: typing.Union[str, None]

    def __init__(self, exit_status: str, exception: typing.Union[str, None] = None):
        if exit_status not in [
            "SUCCESS",
            "LOCK_ACQUISITION_FAILED",
            "GENERIC_EXCEPTION",
        ]:
            raise ValueError(
                f"exit_status must be one of ['SUCCESS', 'LOCK_ACQUISITION_FAILED', 'GENERIC_EXCEPTION'], got {exit_status}"
            )

        self.exit_status = exit_status
        self.exception = exception


class ExecutionIdMessage(Message):
    type = "execution-id"

    def __init__(self, execution_id: str):
        super().__init__({"executionId": execution_id})


class CloseMessage(Message):
    type = "program:end"

    def __init__(self, close_dto: CloseDTO):
        super().__init__(
            {"exitStatus": close_dto.exit_status, "exception": close_dto.exception}
        )


class FormUpdateMessage(Message):
    type = "form-update"

    def __init__(self, widgets: typing.List, validation: typing.Dict, event_seq: int):
        super().__init__(
            {
                "widgets": widgets,
                "validation": validation,
                "seq": event_seq,
            }
        )


class FormMessage(Message):
    type = "form"

    def __init__(
        self,
        widgets: typing.List,
        actions: typing.List,
        end_program: bool,
        reactive_polling_interval: int,
        steps: typing.List,
    ):
        super().__init__(
            {
                "widgets": widgets,
                "actions": actions,
                "endProgram": end_program,
                "reactivePollingInterval": reactive_polling_interval,
                "steps": steps,
            }
        )



import flask

from ...usage import usage
from ...settings import Settings
from .main import MainController


def get_editor_bp(controller: MainController):
    bp = flask.Blueprint("editor_workspace", __name__)

    @bp.get("/")
    @usage
    def _get_workspace():
        return controller.get_workspace().editor_dto

    @bp.put("/")
    @usage
    def _update_workspace():
        if not flask.request.json:
            flask.abort(400)
        workspace = controller.update_workspace(flask.request.json)
        return workspace.editor_dto

    @bp.get("/root")
    @usage
    def _get_workspace_root_path():
        return str(Settings.root_path.absolute())

    @bp.post("/open-file")
    @usage
    def _open_file():
        if not flask.request.json:
            flask.abort(400)
        file_path = flask.request.json["path"]
        mode = flask.request.json.get("mode", "file")

        controller.open_file(file_path, create_if_not_exists=True, mode=mode)
        return {"success": True}

    @bp.get("/check-file")
    @usage
    def _check_file():
        file_path = flask.request.args.get("path")
        if not file_path:
            flask.abort(400)

        return {"exists": controller.check_file(file_path)}

    @bp.get("/files")
    @usage
    def _list_files():
        path = flask.request.args.get("path", ".")
        mode = flask.request.args.get("mode", "file")
        return controller.list_files(path, mode)

    @bp.post("/deploy")
    @usage
    def _deploy():
        controller.deploy()
        return {"success": True}

    return bp



from typing import Union, List

from ..linter import LinterRule, LinterFix, LinterIssue
from ...templates import new_hook_code, new_job_code, new_script_code, new_form_code
from ...repositories.project.project import (
    ProjectRepository,
    ScriptStage,
    FormStage,
    HookStage,
    JobStage,
    WorkflowStage,
)


class AddEntrypoint(LinterFix):
    label = "Add entrypoint"
    description = "Creates the .py file for the entrypoint"
    runtime: Union[FormStage, HookStage, JobStage, ScriptStage]

    def __init__(
        self, runtime: Union[FormStage, HookStage, JobStage, ScriptStage]
    ) -> None:
        self.runtime = runtime

    def make_label(self):
        return f"Create {self.runtime.file}"

    def fix(self):
        if isinstance(self.runtime, FormStage):
            self.runtime.file_path.write_text(new_form_code)
        elif isinstance(self.runtime, HookStage):
            self.runtime.file_path.write_text(new_hook_code)
        elif isinstance(self.runtime, JobStage):
            self.runtime.file_path.write_text(new_job_code)
        elif isinstance(self.runtime, ScriptStage):
            self.runtime.file_path.write_text(new_script_code)
        else:
            raise Exception(f"Unknown runtime: {self.runtime}")

    @property
    def name(self):
        return f"{self.__class__.__name__}:{self.runtime.__class__.__name__}:{self.runtime.id}"


class NoEntrypointFound(LinterIssue):
    def __init__(self, runtime: WorkflowStage) -> None:
        self.label = f"The {runtime.runner_type} entitled {runtime.title} points to a non-existent file: {runtime.file}"
        self.fixes = [AddEntrypoint(runtime)]


class MissingEntrypoint(LinterRule):
    label = "Pointed files should exist"
    type = "bug"

    def find_issues(self) -> List[LinterIssue]:
        project = ProjectRepository.load()
        issues = []
        for form in project.forms:
            if not form.file_path.exists():
                issues.append(NoEntrypointFound(form))

        for hook in project.hooks:
            if not hook.file_path.exists():
                issues.append(NoEntrypointFound(hook))

        for job in project.jobs:
            if not job.file_path.exists():
                issues.append(NoEntrypointFound(job))

        for script in project.scripts:
            if not script.file_path.exists():
                issues.append(NoEntrypointFound(script))

        return issues



import threading, os, time
from datetime import datetime

from ..contract.forms import FilesChangedMessage
from ..modules import reload_project_local_modules
from ..execution.form_execution import FormExecution
from ..repositories.project.project import ProjectRepository, Project


def _has_local_dependencies_changed(project: Project, last_change: float) -> bool:
    for file in project.project_local_dependencies:
        if not file.exists():
            continue

        st_mtime = os.stat(file).st_mtime
        if st_mtime > last_change:
            return True
    return False


def reload_files_on_change(project: Project, last_change: float):
    if _has_local_dependencies_changed(project, last_change):
        reload_project_local_modules()
        FormExecution.broadcast(FilesChangedMessage())
        return True
    else:
        return False


def __files_changed_polling_loop():
    last_change = datetime.now().timestamp()
    project = ProjectRepository.load()

    while True:
        has_reloaded = reload_files_on_change(project, last_change)
        if has_reloaded:
            last_change = datetime.now().timestamp()

        time.sleep(1)


def watch_py_root_files():
    threading.Thread(target=lambda: __files_changed_polling_loop()).start()



import json, os
from pathlib import Path
from datetime import datetime
from typing import List, Type

from .base_migration import Migration
from .migration_001 import Migration001


MIGRATIONS: List[Type[Migration]] = [
    Migration001,
]


def get_latest_version() -> str:
    return MIGRATIONS[-1].target_version()


def create_backup(data: dict, location: Path, filename: str):
    backup_path = os.path.join(location, filename)
    with open(backup_path, "w") as file:
        json.dump(data, file)


def migrate(data: dict, path: Path):
    if "version" not in data:
        raise KeyError("Missing 'version' key in data")

    next_migration = next(
        (m for m in MIGRATIONS if m.source_version() == data["version"]), None
    )

    if not next_migration:
        return data

    print("Your abstra.json is outdated, running migrations...")

    try:
        filename = f"abstra_{datetime.now().strftime('%Y%m%d%H%M%S')}.json.backup"
        create_backup(data, path, filename)
        print(f"Backup file created: {filename}")
    except Exception as e:
        print(f"Failed to create backup file: {e}")
        raise

    while next_migration:
        current_version = data["version"]

        try:
            instance = next_migration(data)
            instance.apply()
            data = instance.data

            for warning in instance.warnings:
                print(warning)

        except Exception as e:
            print(f"An error occurred during migration from {current_version}: {e}")
            raise

        print(f"Upgrade to version {data['version']} complete.")

        next_migration = next_migration = next(
            (m for m in MIGRATIONS if m.source_version() == data["version"]), None
        )

    print("Your abstra.json is up to date ✅")

    return data
